import httpx

from pysirix.sirix import Sirix
from pysirix.database import Database
from pysirix.resource import Resource
from pysirix.json_store import JsonStoreSync, JsonStoreAsync
from pysirix.constants import Insert, DBType, TimeAxisShift
from pysirix.errors import SirixServerError
from pysirix.types import (
    QueryResult,
    Commit,
    Revision,
    InsertDiff,
    ReplaceDiff,
    UpdateDiff,
    DeleteDiff,
    Metadata,
    MetaNode,
)


def sirix_sync(username: str, password: str, client: httpx.Client,) -> Sirix:
    """
    :param username: the username registered with keycloak for this application.
    :param password: the password registered with keycloak for this application.
    :param client: an ``httpx.Client`` instance. You should instantiate the instance with
            the ``base_url`` param as the url for the sirix database.
    """
    s = Sirix(username=username, password=password, client=client,)
    s.authenticate()
    return s


async def sirix_async(
    username: str, password: str, client: httpx.AsyncClient,
) -> Sirix:
    """
    :param username: the username registered with keycloak for this application.
    :param password: the password registered with keycloak for this application.
    :param client: an ``httpx.AsyncClient`` instance. You should instantiate the instance with
            the ``base_url`` param as the url for the sirix database.
    """
    s = Sirix(username=username, password=password, client=client,)
    await s.authenticate()
    return s


__all__ = [
    "sirix_sync",
    "sirix_async",
    "Sirix",
    "SirixServerError",
    "Database",
    "Resource",
    "JsonStoreSync",
    "JsonStoreAsync",
    "Insert",
    "DBType",
    "QueryResult",
    "Commit",
    "Revision",
    "InsertDiff",
    "ReplaceDiff",
    "UpdateDiff",
    "DeleteDiff",
    "Metadata",
    "MetaNode",
    "TimeAxisShift",
]



from typing import Dict, List, Union, Coroutine, Optional

import httpx

from pysirix.sync_client import SyncClient
from pysirix.async_client import AsyncClient
from pysirix.auth import Auth
from pysirix.database import Database

from pysirix.constants import DBType


class Sirix:
    def __init__(
        self,
        username: str,
        password: str,
        client: Union[httpx.Client, httpx.AsyncClient],
    ):
        """
        SirixDB access class.
        This class is the entrypoint for manipulating data with SirixDB.

        :param username: the username registered with keycloak for this application.
        :param password: the password registered with keycloak for this application.
        :param client: the ``httpx.Client`` or ``httpx.AsyncClient`` to use.
        """
        if isinstance(client, httpx.Client):
            self._client = SyncClient(client)
            self._auth = Auth(username, password, client, False)
        else:
            self._client = AsyncClient(client)
            self._auth = Auth(username, password, client, True)

    def authenticate(self):
        """
        Call the authenticate endpoint. Must be called before any other calls are made.
        This is done internally by :py:func:`sirix_sync` or by :py:func:`sirix_async`.
        """
        return self._auth.authenticate()

    def dispose(self):
        """
        Remove the authentication timer.
        """
        self._auth.dispose()

    def database(self, database_name: str, database_type: DBType):
        """
        Returns a :py:class:`Database` instance.

        :param database_name: the name of the database to access.
        :param database_type: the type of the database to access.
        """
        return Database(database_name, database_type, self._client, self._auth)

    def get_info(
        self, resources: bool = True
    ) -> Union[Coroutine, List[Dict[str, str]]]:
        """
        Returns a list of database names and types, and (optionally) a list their resources as well.

        :param resources: whether or not to include resource information
        :return: a ``list`` of ``dicts``, where each ``dict`` has a ``name`` field, a ``type`` field,
                        and (if resources is ``True``) a ``resources`` field (containing a ``list`` of names).
        """
        return self._client.global_info(resources)

    def query(
        self,
        query: str,
        start_result_seq_index: int = None,
        end_result_seq_index: int = None,
    ):
        """
        Execute a custom query on SirixDB.
        Unlike the query method on :py:class:`Resource`, queries executed with this method
        potentially access the entirety of the SirixDB server.
        The ``start_result_seq_index`` and ``end_result_seq_index`` can be used for pagination.

        :param query: the query ``str`` to execute.
        :param start_result_seq_index: the first index of the results from which to return, defaults to first.
        :param end_result_seq_index: the last index of the results to return, defaults to last.
        :return: the query result.
        """
        query_obj = {
            "query": query,
            "startResultSeqIndex": start_result_seq_index,
            "endResultSeqIndex": end_result_seq_index,
        }
        query_obj = {k: v for k, v in query_obj.items() if v}
        return self._client.post_query(query_obj)

    def delete_all(self) -> Union[Coroutine, None]:
        """
        Deletes all databases and resources in the SirixDB server. Be careful!
        """
        return self._client.delete_all()



from enum import Enum
from datetime import datetime
from typing import Union


class Insert(Enum):
    """
    This Enum class defines the possible options for a resource update
    """

    CHILD = "asFirstChild"
    LEFT = "asLeftSibling"
    RIGHT = "asRightSibling"
    REPLACE = "replace"


class DBType(Enum):
    """
    This Enum class defines the possible database (and resource) types
    """

    XML = "application/xml"
    JSON = "application/json"


class TimeAxisShift(Enum):
    oldest = -1
    none = 0
    latest = 1


class MetadataType(Enum):
    """
    This class defines the scope of the metadata to return using the
    :py:meth:`readWithMetadata` method.
    """

    ALL = True
    KEY = "nodeKey"
    KEYAndCHILD = "nodeKeyAndChildCount"


Revision = Union[int, datetime]



from httpx import Client

import xml.etree.ElementTree as ET
from typing import Dict, Union, List

from pysirix.constants import DBType, Insert
from pysirix.errors import include_response_text_in_errors
from pysirix.types import Commit, InsertDiff, ReplaceDiff, UpdateDiff, BytesLike

ET.register_namespace("rest", "https://sirix.io/rest")


class SyncClient:
    def __init__(self, client: Client):
        """
        The methods of this class call all SirixDB endpoints, with minimal handling.
        This class is used for synchronous calls, the :py:class:`AsyncClient` handles asynchronous calls.

        :param client: an instance of ``httpx.Client``.
        """
        self.client = client

    def global_info(self, resources: bool = True) -> List[Dict]:
        """
        Call the ``/`` endpoint with a GET request. If ``resources`` is ``True``,
        the endpoint is called with the query ``withResources=true``

        :param resources: whether to query resources as well
        :return: a ``list`` of ``dict``\s, where each ``dict`` has a ``name``
                        field, a ``type`` field, and (if ``resources`` is
                        ``True``) a ``resources`` field (containing a ``list`` of names).
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        params = {}
        if resources:
            params["withResources"] = True
        resp = self.client.get("/", params=params)
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.json()["databases"]

    def delete_all(self) -> None:
        """
        Call the ``/`` endpoint with DELETE request.
        Deletes all databases and their resources. Be careful!

        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.delete("/")
        with include_response_text_in_errors():
            resp.raise_for_status()

    def create_database(self, name: str, db_type: DBType) -> None:
        """
        Call the ``/{database}`` endpoint with a PUT request

        :param name: name of the database to create.
        :param db_type: type of the database to create.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.put(name, headers={"Content-Type": db_type.value})
        with include_response_text_in_errors():
            resp.raise_for_status()

    def get_database_info(self, name: str) -> Dict:
        """
        Call the ``/{database}`` endpoint with a GET request.

        :param name: name of the database.
        :return: a ``dict`` with a ``resources`` field containing a ``list`` of resources.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.get(name, headers={"Accept": "application/json"})
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.json()

    def delete_database(self, name: str) -> None:
        """
        call the ``/{database}`` endpoint with a DELETE request.

        :param name: the name of the database to delete.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.delete(name)
        with include_response_text_in_errors():
            resp.raise_for_status()

    def resource_exists(self, db_name: str, db_type: DBType, name: str) -> bool:
        """
        Call the ``/{database}/{resource}`` endpoint with a HEAD request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :return: a ``bool`` indicating the existence (or lack thereof) of the resource.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.head(f"{db_name}/{name}", headers={"Accept": db_type.value})
        if resp.status_code == 200:
            return True
        return False

    def create_resource(
        self,
        db_name: str,
        db_type: DBType,
        name: str,
        data: BytesLike,
        hash_type: str = "ROLLING",
    ) -> str:
        """
        Call the ``/{database}/{resource}`` endpoint with a PUT request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :param data: the data to initialize the database with.
        :return: a ``str`` of ``data``.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.put(
            f"{db_name}/{name}",
            headers={"Content-Type": db_type.value},
            params={"hash_type": hash_type},
            content=data,
        )
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.text

    def read_resource(
        self,
        db_name: str,
        db_type: DBType,
        name: str,
        params: Dict[str, Union[str, int]],
    ) -> Union[Dict, List, ET.Element]:
        """
        Call the ``/{database}/{resource}`` endpoint with a GET request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :param params: query parameters to call the endpoint with.
        :return: either a ``dict`` or a ``xml.etree.ElementTree.Element``, depending on the database type.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.get(
            f"{db_name}/{name}", params=params, headers={"Accept": db_type.value}
        )
        with include_response_text_in_errors():
            resp.raise_for_status()
        if db_type == DBType.JSON:
            return resp.json()
        else:
            return ET.fromstring(resp.text)

    def history(self, db_name: str, db_type: DBType, name: str) -> List[Commit]:
        """
        Call the ``/{database}/{resource}/history`` endpoint with a GET request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :return: a ``list`` of ``dict`` containing the history of the resource.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.get(
            f"{db_name}/{name}/history", headers={"Accept": db_type.value}
        )
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.json()["history"]

    def diff(
        self, db_name: str, name: str, params: Dict[str, str]
    ) -> List[Dict[str, Union[InsertDiff, ReplaceDiff, UpdateDiff, int]]]:
        """
        Call the ``/{database}/{resource}/diff`` endpoint with a GET request.

        :param db_name: the name of the database.
        :param name: the name of the resource.
        :param params: the parameters required for this request.
        :return:
        """
        resp = self.client.get(f"{db_name}/{name}/diff", params=params)
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.json()["diffs"]

    def post_query(self, query: Dict[str, Union[int, str]]) -> str:
        """
        Call the ``/`` endpoint with a POST request.

        :param query: the body of the request.
        :return: the query result as a ``str``.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.post("/", json=query)
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.text

    def get_etag(
        self,
        db_name: str,
        db_type: DBType,
        name: str,
        params: Dict[str, Union[str, int]],
    ) -> str:
        """
        Call the ``/{database}/{resource}`` endpoint with a HEAD request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :param params: the query parameters to call the endpoint with.
        :return: the ETag of the node queried.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        resp = self.client.head(
            f"{db_name}/{name}", params=params, headers={"Accept": db_type.value}
        )
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.headers["etag"]

    def update(
        self,
        db_name: str,
        db_type: DBType,
        name: str,
        node_id: int,
        data: BytesLike,
        insert: Insert,
        etag: Union[str, None],
    ) -> str:
        """
        Call the ``/{database}/{resource}`` endpoint with a POST request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :param node_id: the nodeKey of the node in relation to which the update is performed.
        :param data: the data used in the update operation.
        :param insert: the position of the update in relation to the node referenced by node_id.
        :param etag: the ETag of the node referenced by node_id.
        :return: the resource as a ``str``.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        if not etag:
            etag = self.get_etag(db_name, db_type, name, {"nodeId": node_id})
        resp = self.client.post(
            f"{db_name}/{name}",
            params={"nodeId": node_id, "insert": insert.value},
            headers={"ETag": etag, "Content-Type": db_type.value},
            content=data,
        )
        with include_response_text_in_errors():
            resp.raise_for_status()
        return resp.text

    def resource_delete(
        self,
        db_name: str,
        db_type: DBType,
        name: str,
        node_id: Union[int, None],
        etag: Union[str, None],
    ) -> None:
        """
        Call the ``/{database}/{resource}`` endpoint with a DELETE request.

        :param db_name: the name of the database.
        :param db_type: the type of the database.
        :param name: the name of the resource.
        :param node_id: the nodeKey of the node to delete. ``None`` to delete the entire resource.
        :param etag: the etag of the node to delete.
        :raises: :py:class:`pysirix.SirixServerError`.
        """
        if node_id and not etag:
            etag = self.get_etag(db_name, db_type, name, {"nodeId": node_id})
        headers = {"Content-Type": db_type.value}
        if etag:
            headers.update({"ETag": etag})
        params = {}
        if node_id:
            params["nodeId"] = node_id
        resp = self.client.delete(f"{db_name}/{name}", params=params, headers=headers)
        with include_response_text_in_errors():
            resp.raise_for_status()



import json
import xml.etree.ElementTree as ET
from collections.abc import Iterator
from datetime import datetime

from typing import Union, Dict, Tuple, Awaitable, Optional, List

from pysirix.auth import Auth
from pysirix.constants import Insert, Revision, DBType, MetadataType

from pysirix.sync_client import SyncClient
from pysirix.async_client import AsyncClient
from pysirix.types import Commit


class Resource:
    def __init__(
        self,
        db_name: str,
        db_type: DBType,
        resource_name: str,
        client: Union[SyncClient, AsyncClient],
        auth: Auth,
    ):
        """
        Resource access class.

        This class allows for manipulation of a resource

        :param db_name: the name of the database this resource belongs to.
        :param db_type: the type of data the database can hold.
        :param resource_name: the name of the resource being accessed, or to
                be created if the resource does not yet exist
        :param client: the :py:class:`SyncClient` or :py:class:`AsyncClient`
                instance to use for network requests
        :param auth: the :py:class:`Auth` that keeps the client authenticated.
                It is referenced to ensure that it never goes out of scope
        """
        self.db_name = db_name
        self.db_type = db_type
        self.resource_name = resource_name
        self._client = client
        self._auth = auth

    def create(self, data: Union[str, Dict, ET.Element], hash_type: str = "ROLLING"):
        """
        :param data: the data with which to initialize the resource.
                May be an instance of ``dict``, or an instance of
                ``xml.etree.ElementTree.Element`` (depending on the database type),
                or a ``str`` of properly formed json or xml.
        """
        data = (
            data
            if type(data) is str or isinstance(data, Iterator)
            else json.dumps(data)
            if self.db_type == DBType.JSON
            else ET.tostring(data)
        )
        return self._client.create_resource(
            self.db_name,
            self.db_type,
            self.resource_name,
            data,
            hash_type,
        )

    def exists(self):
        """
        Sends a ``head`` request to determine whether or not this store/resource already exists.

        :return: a ``bool`` corresponding to the existence of the store.
        """
        return self._client.resource_exists(
            self.db_name, self.db_type, self.resource_name
        )

    def read(
        self,
        node_id: Union[int, None],
        revision: Union[Revision, Tuple[Revision, Revision], None] = None,
        max_level: Union[int, None] = None,
        top_level_limit: Optional[int] = None,
        top_level_skip_last_node: Optional[int] = None,
    ) -> Union[Union[dict, ET.Element], Awaitable[Union[dict, ET.Element]]]:
        """
        Read the node (and its sub-nodes) corresponding to ``node_id``.

        :param node_id: the nodeKey corresponding to the node to read, if ``None``,
                        the entire resource is read.
        :param revision: the revision to read from, defaults to latest.
        :param max_level: the maximum depth for reading sub-nodes, defaults to latest.
        :param top_level_limit: the maximum number of top level nodes to return (used for paging).
        :param top_level_skip_last_node: the last nodeId to skip (used for paging).
        :return: either a ``dict`` or an instance of ``xml.etree.ElementTree.Element``,
                        depending on the database type of this resource.
        """
        params = self._build_read_params(
            node_id, revision, max_level, top_level_limit, top_level_skip_last_node
        )
        return self._client.read_resource(
            self.db_name, self.db_type, self.resource_name, params
        )

    def read_with_metadata(
        self,
        node_id: Union[int, None],
        revision: Union[Revision, Tuple[Revision, Revision], None] = None,
        meta_type: MetadataType = MetadataType.ALL,
        max_level: Optional[int] = None,
        top_level_limit: Optional[int] = None,
        top_level_skip_last_node: Optional[int] = None,
    ):
        """
        Read the node (and its sub-nodes) corresponding to ``node_id``, with metadata for each node.

        :param node_id: the nodeKey corresponding to the node to read, if ``None``,
                        the entire resource is read.
        :param revision: the revision to read from, defaults to latest.
        :param meta_type: the type of metadata to return, defaults to all.
        :param max_level: the maximum depth for reading sub-nodes, defaults to latest.
        :param top_level_limit: the maximum number of top level nodes to return (used for paging).
        :param top_level_skip_last_node: the last nodeId to skip (used for paging).
        :return:
        """
        params = self._build_read_params(
            node_id, revision, max_level, top_level_limit, top_level_skip_last_node
        )
        params["withMetadata"] = meta_type.value
        return self._client.read_resource(
            self.db_name, self.db_type, self.resource_name, params
        )

    @staticmethod
    def _build_read_params(
        node_id: Union[int, None],
        revision: Union[Revision, Tuple[Revision, Revision], None] = None,
        max_level: Union[int, None] = None,
        top_level_limit: Optional[int] = None,
        top_level_skip_last_node: Optional[int] = None,
    ) -> Dict[str, Union[str, int]]:
        """
        Helper method to build a parameters ``dict`` for reading a resource.
        """
        params = {}
        if node_id:
            params["nodeId"] = node_id
        if max_level:
            params["maxLevel"] = max_level
        if top_level_limit:
            params["nextTopLevelNodes"] = top_level_limit
        if top_level_skip_last_node:
            params["lastTopLevelNodeKey"] = top_level_skip_last_node
        if revision:
            if type(revision) == int:
                params["revision"] = revision
            elif isinstance(revision, datetime):
                params["revision-timestamp"] = revision.isoformat()
            if type(revision) == tuple:
                if isinstance(revision[0], datetime):
                    params["start-revision-timestamp"] = revision[0].isoformat()
                    params["end-revision-timestamp"] = revision[1].isoformat()
                else:
                    params["start-revision"] = revision[0]
                    params["end-revision"] = revision[1]
        return params

    def history(self) -> List[Commit]:
        """
        Get a ``list`` of all commits/revision of this resource.

        :return: a ``list`` of ``dict`` of the form :py:class:`pysirix.Commit`.
        """
        return self._client.history(self.db_name, self.db_type, self.resource_name)

    def diff(
        self,
        first_revision: Revision,
        second_revision: Revision,
        node_id: int = None,
        max_depth: int = None,
    ):
        params = {}
        if isinstance(first_revision, datetime):
            params["first-revision"] = first_revision.isoformat()
        else:
            params["first-revision"] = first_revision
        if isinstance(second_revision, datetime):
            params["second-revision"] = second_revision.isoformat()
        else:
            params["second-revision"] = second_revision
        if node_id is not None:
            params["startNodeKey"] = node_id
        if max_depth is not None:
            params["maxDepth"] = max_depth
        return self._client.diff(self.db_name, self.resource_name, params)

    def get_etag(self, node_id: int) -> Union[str, Awaitable[str]]:
        """
        Get the ETag of a given node.

        :param node_id: the nodeKey corresponding to which the ETag should be returned.
        :return: a ``str`` ETag.
        """
        params = {"nodeId": node_id}
        """
        if revision:
            if type(revision) == int:
                params["revision"] = revision
            else:
                params["revision-timestamp"] = revision.isoformat()
        """
        return self._client.get_etag(
            self.db_name, self.db_type, self.resource_name, params
        )

    def update(
        self,
        node_id: int,
        data: Union[str, ET.Element, Dict],
        insert: Insert = Insert.CHILD,
        etag: str = None,
    ) -> Union[str, Awaitable[str]]:
        """
        Update a resource.

        :param node_id: the nodekey in reference to which the update should be performed.
        :param data: the updated data, can be of type ``str``, ``dict``, or
                ``xml.etree.ElementTree.Element``
        :param insert: the position of the update in relation to the node referenced by node_id.
        :param etag: the ETag of the node referenced by node_id.
        """
        data = (
            data
            if type(data) is str
            else json.dumps(data)
            if self.db_type == DBType.JSON
            else ET.tostring(data)
        )
        return self._client.update(
            self.db_name, self.db_type, self.resource_name, node_id, data, insert, etag
        )

    def query(
        self,
        query: str,
        start_result_seq_index: int = None,
        end_result_seq_index: int = None,
    ):
        """
        Execute a custom query on this resource.
        The ``start_result_seq_index`` and ``end_result_seq_index`` can be used for pagination.

        :param query: the query ``str`` to execute.
        :param start_result_seq_index: the first index of the results from which to return, defaults to first.
        :param end_result_seq_index: the last index of the results to return, defaults to last.
        :return: the query result.
        """
        params = {
            "query": query,
            "startResultSeqIndex": start_result_seq_index,
            "endResultSeqIndex": end_result_seq_index,
        }
        params = {k: v for k, v in params.items() if v}
        return self._client.read_resource(
            self.db_name, self.db_type, self.resource_name, params
        )

    def delete(
        self, node_id: Union[int, None], etag: Union[str, None]
    ) -> Union[None, Awaitable[None]]:
        """
        Delete a node in a resource, or, if ``node_id`` is specified as ``None``,
        delete the entire resource.

        :param node_id: an ``int`` corresponding to the node to delete.
                        Should be specified as none to delete the entire resource.
        :param etag: the ``etag`` of the node to delete. This can be fetched using
                        the py:method`get_etag` method. If ``etag`` is specified as
                        ``None``, then the ``etag`` will be fetched and provided implicitly.
        """
        return self._client.resource_delete(
            self.db_name, self.db_type, self.resource_name, node_id, etag
        )



from enum import Enum


class NodeType(Enum):
    OBJECT = "OBJECT"
    ARRAY = "ARRAY"
    OBJECT_KEY = "OBJECT_KEY"
    OBJECT_STRING_VALUE = "OBJECT_STRING_VALUE"
    STRING_VALUE = "STRING_VALUE"
    OBJECT_NUMBER_VALUE = "OBJECT_NUMBER_VALUE"
    NUMBER_VALUE = "NUMBER_VALUE"
    OBJECT_BOOLEAN_VALUE = "OBJECT_BOOLEAN_VALUE"
    BOOLEAN_VALUE = "BOOLEAN_VALUE"
    OBJECT_NULL_VALUE = "OBJECT_NULL_VALUE"
    NULL_VALUE = "NULL_VALUE"


class InsertPosition(Enum):
    asFirstChild = "asFirstChild"
    asLeftSibling = "asLeftSibling"
    asRightSibling = "asRightSibling"
    replace = "replace"


class DataType(Enum):
    string = "string"
    number = "number"
    boolean = "boolean"
    null = "null"
    jsonFragment = "jsonFragment"



from datetime import datetime
from typing import Union, Dict, List, Awaitable, Optional

from pysirix.types import Commit, Revision as RevisionType, SubtreeRevision
from json import dumps, loads

from abc import ABC

from pysirix.constants import DBType, Revision, TimeAxisShift
from pysirix.async_client import AsyncClient
from pysirix.auth import Auth
from pysirix.sync_client import SyncClient
from pysirix.types import QueryResult


def parse_revision(revision: Revision, params: Dict) -> None:
    if type(revision) == int:
        params["revision"] = revision
    elif isinstance(revision, datetime):
        params["revision-timestamp"] = revision.isoformat()


def stringify(v: Union[None, int, str, Dict, List]):
    return (
        f'"{v}"'
        if isinstance(v, str)
        else f"{v}"
        if isinstance(v, int)
        else "true()"
        if v is True
        else "false()"
        if v is False
        else "jn:null()"
        if v is None
        # not a primitive value
        else f"jn:parse('{dumps(v)}')"
    )


query_function_include = (
    "declare function local:q($i, $q) {"
    "let $m := for $k in jn:keys($q) return if (not(empty($i=>$k))) then deep-equal($i=>$k, $q=>$k) else false()"
    " return empty(index-of($m, false()))"
    "};"
)

upsert_function_include = (
    "declare %updating function local:upsert-fields($r, $u) {"
    "for $key in bit:fields($u) return if (empty($r=>$key)) then insert json $u into $r"
    " else replace json value of $r=>$key with $u=>$key"
    "};"
)

update_function_include = (
    "declare %updating function local:update-fields($r, $u) {"
    "for $key in bit:fields($u) return replace json value of $r=>$key with $u=>$key"
    "};"
)


class JsonStoreBase(ABC):
    __slots__ = ("db_name", "db_type", "name", "_client", "_auth")

    def __init__(
        self,
        db_name: str,
        name: str,
        client: Union[SyncClient, AsyncClient],
        auth: Auth,
        root: str = "",
    ):
        self.db_name = db_name
        self.db_type = DBType.JSON
        self.name = name
        self.root = root
        self._client = client
        self._auth = auth

    def insert_one(self, insert_dict: Union[str, Dict]) -> Union[str, Awaitable[str]]:
        """
        Inserts a single record into the store. New records are added at the tail of the store.

        :param insert_dict: either a JSON string of a ``dict``, or a ``dict`` that can be converted to JSON.
        :return: an emtpy ``str`` or an empty ``Awaitable[str]``.
        """
        insert_dict = dumps(insert_dict)
        query = f"append json jn:parse('{insert_dict}') into jn:doc('{self.db_name}','{self.name}'){self.root}"
        return self._client.post_query({"query": query})

    def insert_many(
        self, insert_list: Union[str, List[Dict]]
    ) -> Union[str, Awaitable[str]]:
        """
        Inserts a list of records into the store. New records are added at the the tail of the store.

        :param insert_list: either a JSON string of ``list`` of ``dict``\s, or a ``list`` that can be converted to JSON
        :return: a ``str`` "{rest: []}" or an ``Awaitable[str]`` resolving to this string.
        """
        insert_list = dumps(insert_list)
        query = (
            f"let $doc := jn:doc('{self.db_name}','{self.name}'){self.root}"
            f"for $i in jn:parse('{insert_list}') return append json $i into $doc"
        )
        return self._client.post_query({"query": query})

    def exists(self) -> Union[bool, Awaitable[bool]]:
        """
        Sends a ``head`` request to determine whether or not this store/resource already exists.

        :return: a ``bool`` corresponding to the existence of the store.
        """
        return self._client.resource_exists(self.db_name, self.db_type, self.name)

    def create(self, data: str = "[]") -> Union[str, Awaitable[str]]:
        """
        Creates the store, will overwrite the store if it already exists.

        :param data: data with which to initialize the store
        :return: will return the string "[]". If in async mode, an awaitable that resolves this string.
        """
        return self._client.create_resource(
            self.db_name,
            self.db_type,
            self.name,
            data,
        )

    def resource_history(self) -> Union[List[Commit], Awaitable[List[Commit]]]:
        """
        This method returns the entire history of a resource.

        :return: a list of :py:class:`Commit`.
        """
        return self._client.history(self.db_name, self.db_type, self.name)

    def history_embed(self, node_key: int, revision: Optional[Revision] = None):
        query = (
            f"let $node := sdb:select-item(., {node_key}) let $result := for $rev in jn:all-times($node)"
            f" return if (not(exists(jn:previous($rev)))) then $rev"
            f" else if (sdb:hash($rev) ne sdb:hash(jn:previous($rev))) then $rev"
            " else () return $result"
        )
        params = {"query": query}
        if revision:
            parse_revision(revision, params)
        return self._client.read_resource(self.db_name, self.db_type, self.name, params)

    def history(
        self, node_key: int, subtree: bool = True, revision: Optional[Revision] = None
    ):
        """
        This method returns the history of a node in the resource.

        :param node_key: the root of the subtree whose history should be returned. Defaults to document root.
        :param subtree: whether to account for changes in the subtree of the given node. Defaults to ``True``.
        :param revision: the revision in which the node with the given ``node_key`` exists
                (if it does not exist currently). Defaults to the latest revision. May be an integer or a
                ``datetime`` instance
        :return: If ``subtree`` is ``True``, a list of :py:class:`pysirix.types.SubtreeRevision`. Else, a list of
                :py:class:`pysirix.types.RevisionType`.
        """
        if subtree:
            revision_data = '{"revisionNumber": sdb:revision($rev), "revisionTimestamp": xs:string(sdb:timestamp($rev))}'
            query = (
                f"let $node := sdb:select-item(., {node_key}) let $result := for $rev in jn:all-times($node)"
                f" return if (not(exists(jn:previous($rev)))) then {revision_data}"
                f" else if (sdb:hash($rev) ne sdb:hash(jn:previous($rev))) then {revision_data}"
                " else () return $result"
            )
        else:
            query = f"sdb:item-history(sdb:select-item(., {node_key}))"
        params = {"query": query}
        if revision:
            parse_revision(revision, params)
        return self._client.read_resource(self.db_name, self.db_type, self.name, params)

    def _prepare_find_all(
        self,
        query_dict: Dict,
        projection: List[str] = None,
        revision: Revision = None,
        node_key: bool = True,
        hash: bool = False,
        time_axis_shift: TimeAxisShift = TimeAxisShift.none,
        start_result_index: Optional[int] = None,
        end_result_index: Optional[int] = None,
    ):
        if revision is None:
            query_list = [
                query_function_include,
                f"for $i in jn:doc('{self.db_name}','{self.name}'){self.root}",
            ]
        elif isinstance(revision, datetime):
            query_list = [
                query_function_include,
                "for $i in bit:array-values(jn:open"
                f"('{self.db_name}','{self.name}',xs:dateTime('{revision.isoformat()}')){self.root})",
            ]
        else:
            query_list = [
                query_function_include,
                f"for $i in bit:array-values(jn:doc('{self.db_name}','{self.name}',{revision})){self.root}",
            ]
        query_list.append(f"where local:q($i, {stringify(query_dict)})")
        if time_axis_shift == TimeAxisShift.oldest:
            query_list.append("let $i := jn:first-existing($i)")
        elif time_axis_shift == TimeAxisShift.latest:
            query_list.append("let $i := jn:last-existing($i)")
        return_obj = (
            "return {$i,'nodeKey': sdb:nodekey($i),'hash': sdb:hash($i)}"
            if node_key and hash
            else "return {$i,'nodeKey': sdb:nodekey($i)}"
            if node_key
            else "return {$i,'hash': sdb:hash($i)}"
            if hash
            else "return {$i}"
        )
        query_list.append(return_obj)
        query_string = " ".join(query_list)
        if projection is not None:
            if node_key:
                projection.append("nodeKey")
            if hash:
                projection.append("hash")
            projection_string = ",".join(projection)
            query_string = "".join([query_string, "{", projection_string, "}"])
        params = {"query": query_string}
        if start_result_index is not None:
            params["startResultSeqIndex"] = start_result_index
        if end_result_index is not None:
            params["endResultSeqIndex"] = end_result_index
        return params

    def find_all(
        self,
        query_dict: Dict,
        projection: List[str] = None,
        revision: Revision = None,
        node_key: bool = True,
        hash: bool = False,
        time_axis_shift: TimeAxisShift = TimeAxisShift.none,
        start_result_index: Optional[int] = None,
        end_result_index: Optional[int] = None,
    ) -> List[QueryResult]:
        """
        Finds and returns all records where the values of ``query_dict`` match the
        corresponding values the record.

        ``projection`` can optionally be used to retrieve only certain fields of the
        matching records.

        By default, the node_key of of each record is returned as a ``nodeKey`` field in the record.
        The ``node_key`` parameter controls this behavior.

        :param query_dict: a ``dict`` with which to query the records.
        :param projection: a ``list`` of field names to return for the matching records.
        :param node_key: a ``bool`` determining whether or not to return a ``nodeKey`` field containing
                        the nodeKey of the record.
        :param hash: a ``bool`` determining whether or not to return a ``hash`` field containing the hash of the record.
        :param revision: the revision to search, defaults to latest. May be an integer or a ``datetime`` instance
        :param time_axis_shift: specify fetching the most or least recent existing revision of the record
        :param start_result_index: index of first result to return.
        :param end_result_index: index of last result to return.
        :return: a ``list`` of :py:class:`QueryResult` records matching the query.
        """
        raise NotImplementedError()

    def update_by_key(
        self,
        node_key: int,
        update_dict: Dict[str, Union[List, Dict, str, int, None]],
        upsert: bool = True,
    ) -> Union[str, Awaitable[str]]:
        """

        :param node_key: the nodeKey of the record to update
        :param update_dict: a dict of keys and matching values to replace in the given record
        :param upsert: whether to insert if the field does not already exist
        :return:
        """
        query = (
            f"{upsert_function_include if upsert else update_function_include}"
            f"let $rec := sdb:select-item(jn:doc('{self.db_name}','{self.name}'),{node_key}) "
            f"return local:{'upsert' if upsert else 'update'}-fields($rec, {stringify(update_dict)})"
        )
        return self._client.post_query({"query": query})

    def update_many(
        self,
        query_dict: Dict,
        update_dict: Dict[str, Union[List, Dict, str, int, None]],
        upsert: bool = True,
    ) -> Union[str, Awaitable[str]]:
        """

        :param query_dict: a ``dict`` of field names and their values to match against
        :param update_dict: a dict of keys and matching values to replace in the selected record
        :param upsert: whether to insert if the field does not already exist
        :return:
        """
        query = (
            f"{query_function_include}{upsert_function_include if upsert else update_function_include}"
            f"for $i in jn:doc('{self.db_name}','{self.name}'){self.root} where local:q($i, {stringify(query_dict)})"
            f" return local:{'upsert' if upsert else 'update'}-fields($i, {stringify(update_dict)})"
        )
        return self._client.post_query({"query": query})

    def delete_fields_by_key(
        self, node_key: int, fields: List[str]
    ) -> Union[str, Awaitable[str]]:
        """

        :param node_key: the nodeKey of the record to delete
        :param fields: the keys of the fields of the record to delete
        :return:
        """
        query = (
            f"let $obj := sdb:select-item(jn:doc('{self.db_name}','{self.name}'),{node_key})"
            f" let $update := {stringify(fields)}"
            f" for $i in $fields return delete json $obj=>$i"
        )
        return self._client.post_query({"query": query})

    def delete_field(
        self, query_dict: Dict, fields: List[str]
    ) -> Union[str, Awaitable[str]]:
        """

        :param query_dict: a ``dict`` of field names and their values to match against
        :param fields: the keys of the fields of the records to delete
        :return:
        """
        query = (
            f"{query_function_include}"
            f"let $records := for $i in jn:doc('{self.db_name}','{self.name}'){self.root}"
            f" where local:q($i, {stringify(query_dict)}) return $i"
            f" let $fields := {stringify(fields)}"
            f" for $i in $fields return delete json $records=>$i"
        )
        return self._client.post_query({"query": query})

    def delete_records(self, query_dict: Dict) -> Union[str, Awaitable[str]]:
        """

        :param query_dict: a ``dict`` of field names and their values to match against
        :return:
        """
        query = (
            f"{query_function_include}"
            f"let $doc := jn:doc('{self.db_name}','{self.name}'){self.root}"
            f" let $m := for $i at $pos in $doc where local:q($i, {stringify(query_dict)}) return $pos - 1"
            " for $i in $m order by $i descending return delete json $doc[[$i]]"
        )
        return self._client.post_query({"query": query})

    def find_by_key(
        self,
        node_key: Union[int, None],
        revision: Union[Revision, None] = None,
    ):
        """

        :param node_key: the nodeKey of the record to read
        :param revision: the revision to search, defaults to latest. May be an integer or a ``datetime`` instance
        :return:
        """
        params = {"nodeId": node_key}
        if revision:
            parse_revision(revision, params)
        return self._client.read_resource(self.db_name, self.db_type, self.name, params)

    def find_one(
        self,
        query_dict: Dict,
        projection: List[str] = None,
        revision: Revision = None,
        node_key: bool = True,
        hash: bool = False,
        time_axis_shift: TimeAxisShift = TimeAxisShift.none,
    ) -> List[QueryResult]:
        """
        This method is the same as :py:meth:`find_many`, except that this method will only return the first result,
                by way of passing ``0`` to that method's ``start_result_index``, and ``end_result_index`` parameters.

        :param query_dict:
        :param projection:
        :param revision:
        :param node_key:
        :param hash:
        :param time_axis_shift:
        :return:
        """
        return self.find_all(
            query_dict,
            projection,
            revision,
            node_key,
            hash,
            time_axis_shift,
            start_result_index=0,
            end_result_index=0,
        )


class JsonStoreSync(JsonStoreBase):
    """
    This class is a convenient abstraction over the resource entities exposed by SirixDB.
    As such, there is no JsonStore on the SirixDB server, only the underlying resource is stored.

    This class is for storing many distinct, JSON objects in a single resource, where the objects/records
    store data of similar type. As such, it's usage parallels a that of a document store, and an object is
    an abstraction similar to a single document in such a store.
    """

    def find_all(
        self,
        query_dict: Dict,
        projection: List[str] = None,
        revision: Revision = None,
        node_key: bool = True,
        hash: bool = False,
        time_axis_shift: TimeAxisShift = TimeAxisShift.none,
        start_result_index: Optional[int] = None,
        end_result_index: Optional[int] = None,
    ) -> List[QueryResult]:
        params = self._prepare_find_all(
            query_dict,
            projection,
            revision,
            node_key,
            hash,
            time_axis_shift,
            start_result_index,
            end_result_index,
        )
        result = self._client.post_query(params)
        return loads(result)["rest"]

    def history(
        self, node_key: int, subtree: bool = True, revision: Optional[Revision] = None
    ) -> Union[List[SubtreeRevision], List[RevisionType],]:
        return super().history(node_key, subtree, revision)["rest"]

    def resource_history(self) -> List[Commit]:
        return super().resource_history()

    def history_embed(
        self, node_key: int, revision: Optional[Revision] = None
    ) -> List[QueryResult]:
        return super().history_embed(node_key, revision)["rest"]

    def find_by_key(
        self,
        node_key: Union[int, None],
        revision: Union[Revision, None] = None,
    ):
        return super().find_by_key(node_key, revision)


class JsonStoreAsync(JsonStoreBase):
    """
    This class is a convenient abstraction over the resource entities exposed by SirixDB.
    As such, there is no JsonStore on the SirixDB server, only the underlying resource is stored.

    This class is for storing many distinct, JSON objects in a single resource, where the objects/records
    store data of similar type. As such, it's usage parallels a that of a document store, and an object is
    an abstraction similar to a single document in such a store.
    """

    async def find_all(
        self,
        query_dict: Dict,
        projection: List[str] = None,
        revision: Revision = None,
        node_key: bool = True,
        hash: bool = False,
        time_axis_shift: TimeAxisShift = TimeAxisShift.none,
        start_result_index: Optional[int] = None,
        end_result_index: Optional[int] = None,
    ) -> List[QueryResult]:
        params = self._prepare_find_all(
            query_dict,
            projection,
            revision,
            node_key,
            hash,
            time_axis_shift,
            start_result_index,
            end_result_index,
        )
        result = await self._client.post_query(params)
        return loads(result)["rest"]

    async def history(
        self, node_key: int, subtree: bool = True, revision: Optional[Revision] = None
    ) -> Union[List[SubtreeRevision], List[RevisionType]]:
        result = await super().history(node_key, subtree, revision)
        return result["rest"]

    async def resource_history(self) -> List[Commit]:
        return await super().resource_history()

    async def history_embed(
        self, node_key: int, revision: Optional[Revision] = None
    ) -> List[QueryResult]:
        result = await super().history_embed(node_key, revision)
        return result["rest"]

    async def find_by_key(
        self,
        node_key: Union[int, None],
        revision: Union[Revision, None] = None,
    ):
        return await super().find_by_key(node_key, revision)



from .. import DBType, Insert, sirix_sync, SirixServerError
import httpx

import json
import cmd
import argparse
import shlex
from datetime import datetime
import getpass
from pathlib import Path

from xml.etree import ElementTree as ET


def delete_parser():
    parser = argparse.ArgumentParser(
        description=f"Delete server/database/resource/node", prog="delete",
    )

    macro_group = parser.add_mutually_exclusive_group()
    macro_group.add_argument(
        "-s",
        "--server",
        action="store_true",
        help="delete all databases and their resources",
    )
    macro_group.add_argument(
        "-d",
        "--database",
        action="store_true",
        help="delete the entire database in current context",
    )
    macro_group.add_argument(
        "-r",
        "--resource",
        action="store_true",
        help="delete the entire resource in current context",
    )

    node_group = parser.add_argument_group()
    node_group.add_argument(
        "-n", "--nodekey", type=int, help="nodekey of node to delete."
    )
    node_group.add_argument("-e", "--etag", help="hash of node to delete. Optional.")
    return parser


def read_parser():
    parser = argparse.ArgumentParser(description="Read from the resource.", prog="read")
    parser.add_argument(
        "-n",
        "--nodekey",
        type=int,
        help="the nodekey to read from the resource. Defaults to root.",
    )
    parser.add_argument(
        "-r",
        "--revision",
        help="Revision number, or timestamp, to read from. Defaults to latest. For multiple revisions, see --revision-start",
    )
    parser.add_argument(
        "-s",
        "--start-revision",
        help="Revision number, or timestamp, to begin reading from.",
    )
    parser.add_argument(
        "-e",
        "--end-revision",
        help="Revision number, or timestamp, to read from last.",
    )
    parser.add_argument(
        "-d",
        "--max-depth",
        type=int,
        default=None,
        help="The maximum depth to read to. Defaults to maximum.",
    )
    parser.add_argument(
        "-m",
        "--metadata",
        type=bool,
        default=False,
        help="Whether to read with metadata of each node. Defaults to False.",
    )
    parser.add_argument(
        "-j",
        "--skip",
        type=int,
        default=None,
        help="The last nodeId to skip (assuming you want to skip...). Defaults to None.",
    )
    parser.add_argument(
        "-l",
        "--limit",
        type=int,
        default=None,
        help="Limits the number of top level nodes returned. Defaults to None.",
    )
    return parser


def diff_parser():
    parser = argparse.ArgumentParser(description="Read from the resource.", prog="diff")
    parser.add_argument(
        "-s",
        "--start-revision",
        required=True,
        help="Revision number, or timestamp, to begin reading from.",
    )
    parser.add_argument(
        "-e",
        "--end-revision",
        required=True,
        help="Revision number, or timestamp, to read from last.",
    )
    parser.add_argument(
        "-n",
        "--nodekey",
        type=int,
        help="the nodekey to read from the resource. Defaults to root.",
    )
    parser.add_argument(
        "-d",
        "--max-depth",
        type=int,
        default=None,
        help="The maximum depth to read to. Defaults to maximum.",
    )
    return parser


def update_parser():
    parser = argparse.ArgumentParser(
        description=f"Delete server/database/resource/node", prog="update",
    )
    parser.add_argument(
        "-n", "--nodekey", required=True, type=int, help="nodekey of node to delete.",
    )
    parser.add_argument("-e", "--etag", help="hash of node to delete. Optional.")
    data_group = parser.add_mutually_exclusive_group(required=True)
    data_group.add_argument("-d", "--data", help="the data to insert.")
    data_group.add_argument("-p", "--data-path", help="the path to data to insert.")
    parser.add_argument(
        "-i",
        "--insert",
        help="the position to insert the data in, relative to the nodekey. "
        "options (case insensitive) are CHILD, LEFT, RIGHT, REPLACE. defaults to CHILD.",
    )
    return parser


def rquery_parser():
    parser = argparse.ArgumentParser(description="Read from the resource.", prog="diff")
    parser.add_argument(
        "-s", "--start-index", help="Start result index, for pagination.",
    )
    parser.add_argument(
        "-e", "--end-revision", help="End result index, for pagination.",
    )
    parser.add_argument(
        "-q", "--query", required=True, help="the query to execute on the resource.",
    )
    return parser


def query_parser():
    parser = argparse.ArgumentParser(
        description="Execute an unbound custom query context of the server.",
        prog="diff",
    )
    parser.add_argument(
        "-s", "--start-index", help="Start result index, for pagination.",
    )
    parser.add_argument(
        "-e", "--end-revision", help="End result index, for pagination.",
    )
    parser.add_argument(
        "-q", "--query", required=True, help="the query to execute on the resource.",
    )
    return parser


def parse_revision(revision: str):
    try:
        return int(revision)
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%d")
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%dT%H:%M")
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%d %H:%M")
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%dT%H:%M:%S")
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%d %H:%M:%S")
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError:
        pass
    try:
        return datetime.strptime(revision, "%Y-%m-%d %H:%M:%S.%f")
    except ValueError:
        pass
    return None


profile = "empty"


def get_config():
    try:
        home = str(Path.home())
        with open(f"{home}/pysirix-shell/config.json") as f:
            config = json.load(f)
    except FileNotFoundError:
        config = {}
    print("profiles:")
    for k, v in config.items():
        print(json.dumps({k: v}, indent=2))
    print()
    if len(config) == 1:
        selected = next(iter(config))
        print(f"Selecting profile '{selected}'")
    else:
        while True:
            selected = input("please enter your profile: ")
            if selected in config:
                break
    print()
    global profile
    profile = selected
    return dict(**config[selected])


class SirixShell(cmd.Cmd):
    prompt = "PySirix> "

    def preloop(self):
        sirix_uri = config.get("url")
        cert_path = config.get("cert")
        username = config.get("username")
        timeout = config.get("timeout", 5)

        if cert_path is None:
            cert_path = True
        if sirix_uri is None:
            sirix_uri = "https://localhost:9443"
        if username is None:
            username = input(
                "This profile does not have a username set. Enter your username: "
            )
        password = getpass.getpass("enter your password: ")

        self.client = httpx.Client(
            base_url=sirix_uri, verify=cert_path, timeout=timeout
        )
        self.sirix = sirix_sync(username, password, self.client)
        self.context = {"root": profile}
        self.file = None

    def emptyline(self):
        pass

    def do_repeat(self, args):
        """
        Execute the last command entered again.
        """
        super().emptyline()

    def do_last(self, args):
        """
        Print the last result returned by the shell.
        """
        try:
            print(self.last_result)
        except AttributeError:
            print("No results cached yet.")

    def postcmd(self, stop, line):
        print("")
        return super().postcmd(stop, line)

        # ----- record and playback -----

    def do_record(self, arg):
        """
        Save future commands to filename: record <filename>.
        Terminate recording with `record stop`.
        """
        if arg == "stop":
            self._close()
        else:
            home = str(Path.home())
            self.file = open(f"{home}/pysirix-shell/{arg}", "w")

    def do_run(self, arg):
        """
        Run commands from a file:  playback <filename>.
        """
        self._close()
        try:
            home = str(Path.home())
            with open(f"{home}/pysirix-shell/{arg}") as f:
                for command in f:
                    print(f"> {command}")
                    self.cmdqueue.append(command)
        except FileNotFoundError as e:
            print(e)

    def precmd(self, line: str):
        line = line.lower()
        if (
            self.file
            and not line.startswith("playback")
            and not line.startswith("record stop")
        ):
            print(line, file=self.file)
        return line

    def _close(self):
        if self.file:
            self.file.close()
            self.file = None

        # ----- end record and playback -----

    def do_root(self, args):
        """
        Change context to root of sirix server.
        """
        if self.context.get("database"):
            del self.context["database"]
        if self.context.get("resource"):
            del self.context["resource"]
        self.prompt = "PySirix> "

    def do_database(self, args: str):
        """
        Change context to database of <name> and <type>.
        USAGE: "database <name> <type>".
        """
        try:
            name, db_type = shlex.split(args)
        except ValueError:
            print('Error: You must provide two arguments. "<name> <type>".')
            return
        self._database_context(name, db_type)

    def _database_context(self, name: str, database_type: str):
        if database_type.lower() in ["json", "xml"]:
            db_type: DBType = DBType.JSON if database_type.lower() == "json" else DBType.XML
            self.database = self.sirix.database(name, db_type)
            self.context["database"] = (name, db_type)
            if self.context.get("resource"):
                del self.context["resource"]
            self.prompt = f"{self.context['root']}/{name} ({database_type.upper()})> "
        else:
            print(
                'Error: Invalid database type. valid types are "JSON" and "XML" (case insensitive)'
            )

    def do_resource(self, args: str):
        """
        Change context context to resource of <name>.
        USAGE: "resource <name>". Or "resource <db_name> <type> <name>".
        """
        args = shlex.split(args)
        if len(args) == 3:
            self._database_context(args[0], args[1])
            name = args[2]
        elif len(args) != 1:
            print("You must provide 1 or 3 arguments.")
            return
        else:
            name = args[0]
        self._resource_context(name)

    def _resource_context(self, name: str):
        self.resource = self.database.resource(name)
        self.context["resource"] = name
        db_name, db_type = self.context["database"]
        show_type = "JSON" if db_type == DBType.JSON else "XML"
        self.prompt = f"{self.context['root']}/{db_name}/{name} ({show_type})> "

    def do_info(self, args):
        """
        Retrieve information about the servers databases and resources, or
        about the resources of the database in current context. Type `info help`
        for more information.
        """
        if self.context.get("database"):
            try:
                result = self.database.get_database_info()
            except SirixServerError as e:
                print(e)
                return
        else:
            resources = True
            if args in ["none", "false", "no-resources"]:
                resources = False
            try:
                result = self.sirix.get_info(resources)
            except SirixServerError as e:
                print(e)
                return
        print(json.dumps(result, indent=2))
        self.last_result = result

    def do_exists(self, args):
        """
        Returns the existence of the resource in current context.
        """
        if self.context.get("resource"):
            try:
                result = self.resource.exists()
                print(f"exists: {result}")
                self.last_result = result
            except SirixServerError as e:
                print(e)
        else:
            print("Error: Not in context of resource.")

    def do_create(self, args: str):
        """
        Create a database or resource, depending on current context. Type `create help` for more information.
        """
        if self.context.get("resource"):
            parser = argparse.ArgumentParser(
                description=f"Create resource <{self.context['resource']}>",
                prog="create",
            )
            group = parser.add_mutually_exclusive_group(required=True)
            group.add_argument("-p", "--data-path")
            group.add_argument(
                "-d", "--data",
            )
            try:
                parsed = vars(parser.parse_args(shlex.split(args)))
            except:
                return
            if parsed.get("data"):
                data = parsed["data"]
            else:
                with open(parsed["data_path"]) as f:
                    data = f.read()
            try:
                self.resource.create(data)
            except SirixServerError as e:
                print(e)
                return
            print(f"created resource {self.context['resource']}")
        elif self.context.get("database"):
            try:
                self.database.create()
            except SirixServerError as e:
                print(e)
                return
            print(f"created database {self.context['database'][0]}")
        else:
            print("Error: Not in context of database or resource.")

    def do_delete(self, args: str):
        """
        Delete everything on the server, or the database in current context,
        or the resource in current context, or a particular node in the resource
        in current context. Type `delete help` for more information.
        """
        parser = delete_parser()
        try:
            parsed = vars(parser.parse_args(args.split()))
        except:
            return
        try:
            db_name, db_type = self.context.get("database")
        except TypeError:
            db_name, db_type = None, None
        show_type = (
            "JSON"
            if db_type == DBType.JSON
            else "XML"
            if db_type == DBType.XML
            else None
        )
        if parsed.get("nodekey") is not None:
            nodeKey = parsed["nodekey"]
            etag = parsed.get("etag")
            try:
                self.resource.delete(nodeKey, etag)
            except SirixServerError as e:
                print(e)
                return
            print(f"Node with nodeKey: {nodeKey} deleted")
        elif parsed.get("resource"):
            if self.context.get("resource"):
                try:
                    self.resource.delete(None, None)
                except SirixServerError as e:
                    print(e)
                    return
                print(
                    " ".join(
                        [
                            "Deleted resource",
                            f"<{self.context['resource']}>",
                            f"of database <{db_name}> of type <{show_type}>",
                        ]
                    )
                )
            else:
                print("Error: Not in the context of a resource.")
        elif parsed.get("database"):
            if db_name is not None:
                try:
                    self.database.delete()
                    print(f"Deleted database <{db_name}> of type <{show_type}>.")
                except SirixServerError as e:
                    print(e)
            else:
                print("Error: Not in the context of a database.")
        elif parsed.get("server"):
            try:
                self.sirix.delete_all()
                print("Deleted all databases and resources.")
            except SirixServerError as e:
                print(e)
        else:
            print("Error: invalid delete specification")

    def do_read(self, args):
        """
        Read from the resource in current context. Type `read help` for more information.
        """
        if not self.context.get("resource"):
            print("Error: Not in the context of a resource.")
            return
        parser = read_parser()
        try:
            parsed = vars(parser.parse_args(shlex.split(args)))
        except:
            return
        node_key = parsed.get("nodekey")
        max_level = parsed.get("max_depth")
        revision = parsed.get("revision")
        start_revision = parsed.get("start_revision")
        end_revision = parsed.get("end_revision")
        top_level_limit = parsed.get("limit")
        top_level_skip_last_node = parsed.get("skip")
        rev = None
        if start_revision is not None:
            start_rev = parse_revision(start_revision)
            if start_rev is None:
                print(f"Error: {start_revision} is of invalid form.")
                return
            end_rev = parse_revision(end_revision)
            if end_rev is None:
                print(f"Error: {end_revision} is of invalid form.")
                return
            rev = (start_rev, end_rev)
        if revision is not None:
            rev = parse_revision(revision)
            if rev is None:
                print(f"Error: {revision} is of invalid form.")
                return
        try:
            if parsed.get("metadata"):
                result = self.resource.read_with_metadata(
                    node_key,
                    rev,
                    max_level=max_level,
                    top_level_limit=top_level_limit,
                    top_level_skip_last_node=top_level_skip_last_node,
                )
            else:
                result = self.resource.read(
                    node_key, rev, max_level, top_level_limit, top_level_skip_last_node,
                )
            if self.context["database"][1] == DBType.JSON:
                print(json.dumps(result, indent=2))
            else:
                result = ET.tostring(result, encoding="unicode").replace("\\n", "\n")
                print(result)
            self.last_result = result
        except SirixServerError as e:
            print(e)

    def do_history(self, args: str):
        """
        Retrieve history of resource in current context.
        """
        if self.context.get("resource") is None:
            print("Not in the context of a resource.")
            return
        try:
            result = self.resource.history()
            print(json.dumps(result, indent=2))
            self.last_result = result
        except SirixServerError as e:
            print(e)

    def do_hash(self, args):
        """
        Retrieve hash of the node corresponding to <nodekey> in the resource in current context.
        USAGE: hash <nodekey>
        """
        if self.context.get("resource"):
            try:
                node_id = int(args)
            except ValueError:
                print("Error: Invalid nodekey.")
                return
            try:
                result = self.resource.get_etag(node_id)
                self.last_result = result
                print(result)
            except SirixServerError as e:
                print(e)
        else:
            print("Error: Not in the context of a resource.")

    def do_diff(self, args):
        """
        Return the differential between two revision of the resource in
        current context. Type `diff help` for more information.
        """
        if self.context.get("resource") is None:
            print("Not in the context of a resource.")
            return
        parser = diff_parser()
        try:
            parsed = vars(parser.parse_args(shlex.split(args)))
        except:
            return
        start_revision = parsed["start_revision"]
        end_revision = parsed["end_revision"]
        nodekey = parsed.get("nodekey")
        depth = parsed.get("max_depth")
        start_rev = None
        end_rev = None
        start_rev = parse_revision(start_revision)
        if start_rev is None:
            print(f"Error: {start_revision} is of invalid form.")
        end_rev = parse_revision(end_revision)
        if end_rev is None:
            print(f"Error: {end_revision} is of invalid form.")
        try:
            result = self.resource.diff(start_rev, end_rev, nodekey, depth)
            print(json.dumps(result, indent=2))
            self.last_result = result
        except SirixServerError as e:
            print(e)

    def do_update(self, args):
        """
        Update the data in the resource in current context. Type `update help`
        for more information.
        """
        if self.context.get("resource") is None:
            print("Not in the context of a resource.")
            return
        parser = update_parser()
        try:
            parsed = vars(parser.parse_args(shlex.split(args)))
        except:
            return
        nodekey = parsed["nodekey"]
        data = parsed.get("data")
        if data is None:
            with open(parsed["data_path"]) as f:
                data = f.read()
        etag = parsed.get("etag")
        insert = parsed.get("insert")
        if insert is None:
            insert = Insert.CHILD
        else:
            insert = Insert[insert.upper()]
        try:
            self.resource.update(nodekey, data, insert, etag)
        except SirixServerError as e:
            print(e)

    def do_rquery(self, args):
        """
        Execute custom query on a resource. Type `rquery help` for more information.
        """
        parser = rquery_parser()
        try:
            parsed = vars(parser.parse_args(shlex.split(args)))
        except:
            return
        try:
            result = self.resource.query(
                parsed["query"], parsed.get("start_index"), parsed.get("end_index")
            )
            if self.context["database"][1] == DBType.JSON:
                print(json.dumps(result, indent=2))
            else:
                result = ET.tostring(result, encoding="unicode").replace("\\n", "\n")
                print(result)
            self.last_result = result
        except SirixServerError as e:
            print(e)

    def do_query(self, args):
        parser = query_parser()
        try:
            parsed = vars(parser.parse_args(shlex.split(args)))
        except:
            return
        try:
            print(parsed)
            print()
            result = self.sirix.query(
                parsed["query"], parsed.get("start_index"), parsed.get("end_index")
            )
            if self.context["database"][1] == DBType.JSON:
                print(json.dumps(result, indent=2))
            else:
                result = ET.tostring(result, encoding="unicode").replace("\\n", "\n")
                print(result)
            self.last_result = result
        except SirixServerError as e:
            print(e)


def main():
    global config
    config = get_config()
    shell = SirixShell()
    try:
        shell.cmdloop()
    except BaseException as e:
        print(e)
        shell.client.close()
        exit()



from asyncio import ensure_future, sleep
from threading import Timer

import httpx

from typing import Union, Awaitable


class Auth:
    """
    This class handles authentication for server access.
    """

    def __init__(
        self,
        username: str,
        password: str,
        client: Union[httpx.Client, httpx.AsyncClient],
        asynchronous: bool,
    ):
        """
        :param username: the username for this application.
        :param password: the password for this application.
        :param client: the ``httpx.Client`` or ``httpx.AsyncClient``
                        instance used for connecting to the server.
        :param asynchronous: whether or not this application is asynchronous.
        """
        self._username = username
        self._password = password
        self._asynchronous = asynchronous
        self._client = client
        self._refresh_check = True

    def authenticate(self) -> Union[None, Awaitable[None]]:
        """
        Initial authentication for server access, using username and password.
        Access tokens are renewed in the background.
        """
        if self._asynchronous:
            return self._async_authenticate()
        else:
            self._authenticate()

    def dispose(self):
        """
        Remove the authentication timer.
        """
        self._timer.cancel()

    def _authenticate(self):
        """
        Initial authentication, for synchronous, threaded applications.
        """
        resp = self._client.post(
            "/token",
            json={
                "username": self._username,
                "password": self._password,
                "grant_type": "password",
            },
        )
        resp.raise_for_status()
        self._handle_data(resp)

    async def _async_authenticate(self):
        """
        Initial authentication, for asynchronous applications.
        """
        resp = await self._client.post(
            "/token",
            json={
                "username": self._username,
                "password": self._password,
                "grant_type": "password",
            },
        )
        resp.raise_for_status()
        await self._async_handle_data(resp)

    def _refresh(self):
        """
        Refresh the access token, using the refresh token. For synchronous, threaded applications.
        """
        resp = self._client.post(
            "/token", json={"refresh_token": self._token_data['refresh_token']}
        )
        self._handle_data(resp)

    async def _async_refresh(self):
        """
        Refresh the access token, using the refresh token. For asynchronous applications.
        """
        resp = await self._client.post(
            "/token", json={"refresh_token": self._token_data['refresh_token']}
        )
        await self._async_handle_data(resp)

    def _handle_data(self, resp):
        """
        Parse token data, and set a ``threading.Timer`` to refresh the access token again before it expires.

        :param resp: the ``httpx.Response`` object.
        """
        data = {k.replace("-", "_"): v for k, v in resp.json().items()}
        self._token_data = data
        self._client.headers[
            "Authorization"
        ] = f"{self._token_data['token_type']} {self._token_data['access_token']}"
        self._timer = Timer(self._token_data['expires_in'] - 10, self._refresh)
        self._timer.daemon = True
        self._timer.start()

    async def _async_handle_data(self, resp):
        """
        Parse token data, and create an asynchronous task to refresh the access token again before it expires.

        :param resp: the ``httpx.Response`` object.
        """
        data = {k.replace("-", "_"): v for k, v in resp.json().items()}
        self._token_data = data
        self._client.headers[
            "Authorization"
        ] = f"{self._token_data['token_type']} {self._token_data['access_token']}"
        self._timer = ensure_future(self._sleep_then_refresh())

    async def _sleep_then_refresh(self):
        """
        Helper function for :py:func:`_async_handle_data`.
        This method sleeps, then calls :py:func:`_async_refresh`
        10 seconds before the access token is set to expire.
        """
        await sleep(self._token_data['expires_in'] - 10)
        await self._async_refresh()



from contextlib import contextmanager
import httpx


@contextmanager
def include_response_text_in_errors():
    try:
        yield
    except httpx.HTTPStatusError as exc:
        response = exc.response
        message = f"{exc}\nSirixDB error message: {response.text}"
        raise SirixServerError(message, response=response, request=exc.request) from None


class SirixServerError(httpx.HTTPStatusError):
    pass
